"""System, panel, module, permit, and clock attribute definitions."""

from .constants import (
    ACT_ATTR,
    ENABLE_ATTR,
    HNAME_ATTR,
    LISTORD_ATTR,
    MODE_ATTR,
    PARENT_ATTR,
    PROPNAME_ATTR,
    SHOMNU_ATTR,
    SNAME_ATTR,
    SOURCE_ATTR,
    STATIC_ATTR,
    STATUS_ATTR,
    SUBTYP_ATTR,
    TIMOUT_ATTR,
    VACFLO_ATTR,
    VER_ATTR,
)

# System attributes (unique instance)
SYSTEM_ATTRIBUTES = {
    ACT_ATTR,  # ON/OFF but not sure what it does
    "ADDRESS",  # Pool Address
    "AVAIL",  # ON/OFF but not sure what it does
    "CITY",  # Pool City
    "COUNTRY",  # Country obviously (example 'United States')
    "EMAIL",  # primary email for the owner
    "EMAIL2",  # secondary email for the owner
    "HEATING",  # ON/OFF: Pump On During Heater Cool-Down Delay
    HNAME_ATTR,  # same as objnam
    "LOCX",  # (float) longitude
    "LOCY",  # (float) latitude
    "MANHT",  # ON/OFF: Manual Heat
    MODE_ATTR,  # unit system, 'METRIC' or 'ENGLISH'
    "NAME",  # name of the owner
    "PASSWRD",  # a 4 digit password or ''
    "PHONE",  # primary phone number for the owner
    "PHONE2",  # secondary phone number for the owner
    PROPNAME_ATTR,  # name of the property
    "SERVICE",  # 'AUTO' for automatic
    SNAME_ATTR,  # a crazy looking string I assume to be unique to this system
    "START",  # almost looks like a date but no idea
    "STATE",  # Pool State
    STATUS_ATTR,  # ON/OFF
    "STOP",  # same value as START
    "TEMPNC",  # ON/OFF
    "TIMZON",  # (int) Time Zone (example '-8' for US Pacific)
    VACFLO_ATTR,  # ON/OFF, vacation mode
    "VACTIM",  # ON/OFF
    "VALVE",  # ON/OFF: Pump Off During Valve Action
    VER_ATTR,  # (str) software version
    "ZIP",  # Pool Zip Code
}

# System clock attributes
# Note: there are 2 clocks in the system
# one only contains the SOURCE attribute
# the other everything but SOURCE
SYSTIM_ATTRIBUTES = {
    "CLK24A",  # clock mode, 'AMPM' or 'HR24'
    "DAY",  # in 'MM,DD,YY' format
    "DLSTIM",  # ON/OFF, ON for following DST
    HNAME_ATTR,  # same as objnam
    "LOCX",  # (float) longitude
    "LOCY",  # (float) latitude
    "MIN",  # in 'HH,MM,SS' format (24h clock)
    SNAME_ATTR,  # unused really, likely equals to OBJNAM
    SOURCE_ATTR,  # set to URL if time is from the internet
    STATIC_ATTR,  # (ON/OFF) not sure, only seen 'ON'
    "TIMZON",  # (int) timezone (example '-8' for US Pacific)
    "ZIP",  # ZipCode
}

# Panel attributes
PANEL_ATTRIBUTES = {
    HNAME_ATTR,  # equals to OBJNAM
    LISTORD_ATTR,  # (int) used to order in UI
    "OBJLIST",  # [ (objnam) ] the elements managed by the panel
    "PANID",  # ??? only seen 'SHARE'
    SNAME_ATTR,  # friendly name
    STATIC_ATTR,  # only seen 'ON'
    SUBTYP_ATTR,  # only seen 'OCP'
}

# Module attributes
MODULE_ATTRIBUTES = {
    "CIRCUITS",  # [ objects ] the objects that the module controls
    PARENT_ATTR,  # (objnam) the parent (PANEL) of the module
    "PORT",  # (int) no idea
    SNAME_ATTR,  # friendly name
    STATIC_ATTR,  # (ON/OFF) 'ON'
    SUBTYP_ATTR,  # type of the module (like 'I5P' or 'I8PS')
    VER_ATTR,  # (str) the version of the firmware for this module
}

# User/permit attributes
PERMIT_ATTRIBUTES = {
    ENABLE_ATTR,  # (ON/OFF) ON if user is enabled
    "PASSWRD",  # 4 digit code or ''
    SHOMNU_ATTR,  # privileges associated with this user
    SNAME_ATTR,  # friendly name
    STATIC_ATTR,  # (ON/OFF) only seen ON
    SUBTYP_ATTR,  # ADV for administrator, BASIC for guest
    TIMOUT_ATTR,  # (int) in minutes, timeout for user session
}



"""Controller classes for Pentair IntelliCenter.

This module provides controller classes that manage communication
with the Pentair IntelliCenter system using modern asyncio patterns.

Classes:
    ICBaseController: Basic connection and command handling
    ICModelController: Extends ICBaseController with PoolModel management
    ICConnectionHandler: Manages reconnection with exponential backoff
"""

from __future__ import annotations

import asyncio
import contextlib
import logging
import time
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any, ClassVar, Protocol, runtime_checkable

from .attributes import (
    BODY_TYPE,
    CHEM_TYPE,
    CIRCUIT_TYPE,
    HEATER_TYPE,
    LOTMP_ATTR,
    MODE_ATTR,
    OBJTYP_ATTR,
    PARENT_ATTR,
    PROPNAME_ATTR,
    PUMP_TYPE,
    SCHED_TYPE,
    SENSE_TYPE,
    SNAME_ATTR,
    STATUS_ATTR,
    STATUS_OFF,
    STATUS_ON,
    SUBTYP_ATTR,
    SUPER_ATTR,
    SYSTEM_TYPE,
    VER_ATTR,
    HeaterType,
)
from .connection import ICConnection
from .exceptions import ICCommandError, ICConnectionError, ICResponseError

if TYPE_CHECKING:
    from collections.abc import Callable

    from .model import PoolModel
    from .types import ObjectEntry

_LOGGER = logging.getLogger(__name__)

# Configuration constants
MAX_ATTRIBUTES_PER_QUERY = 50  # Maximum attributes per query batch


@dataclass
class ICConnectionMetrics:
    """Tracks connection metrics for observability."""

    requests_sent: int = 0
    requests_completed: int = 0
    requests_failed: int = 0
    reconnect_attempts: int = 0
    successful_connects: int = 0

    def to_dict(self) -> dict[str, int]:
        """Return metrics as a dictionary."""
        return asdict(self)

    def __repr__(self) -> str:
        return (
            f"ICConnectionMetrics(sent={self.requests_sent}, "
            f"completed={self.requests_completed}, failed={self.requests_failed})"
        )


class ICSystemInfo:
    """Represents system information from IntelliCenter.

    Contains metadata like software version, temperature units,
    and a unique identifier.
    """

    ATTRIBUTES_LIST: ClassVar[list[str]] = [
        PROPNAME_ATTR,
        VER_ATTR,
        MODE_ATTR,
        SNAME_ATTR,
    ]

    def __init__(self, objnam: str, params: dict[str, Any]) -> None:
        # Lazy import to avoid loading hashlib at module level
        from hashlib import blake2b

        self._objnam = objnam
        self._prop_name: str = params[PROPNAME_ATTR]
        self._sw_version: str = params[VER_ATTR]
        self._mode: str = params[MODE_ATTR]

        # Generate unique ID from system name
        h = blake2b(digest_size=8)
        h.update(params[SNAME_ATTR].encode())
        self._unique_id = h.hexdigest()

    def __repr__(self) -> str:
        return (
            f"ICSystemInfo(objnam={self._objnam!r}, prop_name={self._prop_name!r}, "
            f"version={self._sw_version!r}, metric={self.uses_metric})"
        )

    @property
    def prop_name(self) -> str:
        """Return the property name."""
        return self._prop_name

    @property
    def sw_version(self) -> str:
        """Return the software version."""
        return self._sw_version

    @property
    def uses_metric(self) -> bool:
        """Return True if system uses metric units."""
        return self._mode == "METRIC"

    @property
    def unique_id(self) -> str:
        """Return unique identifier for this system."""
        return self._unique_id

    @property
    def objnam(self) -> str:
        """Return the object name."""
        return self._objnam

    def update(self, updates: dict[str, Any]) -> None:
        """Update system info from attribute changes."""
        if PROPNAME_ATTR in updates:
            self._prop_name = updates[PROPNAME_ATTR]
        if VER_ATTR in updates:
            self._sw_version = updates[VER_ATTR]
        if MODE_ATTR in updates:
            self._mode = updates[MODE_ATTR]


def prune(obj: Any) -> Any:
    """Remove undefined parameters (where key == value) from object tree."""
    if isinstance(obj, list):
        return [prune(item) for item in obj]
    if isinstance(obj, dict):
        return {k: prune(v) for k, v in obj.items() if k != v}
    return obj


@dataclass
class _RequestContext:
    """Context for tracking a single request's metrics."""

    metrics: ICConnectionMetrics
    success: bool = field(default=False, init=False)

    def __enter__(self) -> _RequestContext:
        self.metrics.requests_sent += 1
        return self

    def __exit__(self, exc_type: type | None, exc_val: Exception | None, exc_tb: Any) -> None:
        if exc_type is None:
            self.metrics.requests_completed += 1
        else:
            self.metrics.requests_failed += 1


class ICBaseController:
    """Controller for communicating with IntelliCenter.

    Uses modern asyncio streams for clean, efficient communication.
    """

    def __init__(
        self,
        host: str,
        port: int = 6681,
        keepalive_interval: float | None = None,
    ) -> None:
        """Initialize the controller.

        Args:
            host: IP address or hostname of IntelliCenter
            port: TCP port (default: 6681)
            keepalive_interval: Seconds between keepalive requests
        """
        self._host = host
        self._port = port
        self._keepalive_interval = keepalive_interval or 90.0

        # Connection
        self._connection: ICConnection | None = None
        self._system_info: ICSystemInfo | None = None

        # Callbacks
        self._disconnected_callback: Callable[[ICBaseController, Exception | None], None] | None = (
            None
        )

        # Metrics
        self._metrics = ICConnectionMetrics()

    def __repr__(self) -> str:
        return (
            f"ICBaseController(host={self._host!r}, port={self._port}, connected={self.connected})"
        )

    @property
    def host(self) -> str:
        """Return the host address."""
        return self._host

    @property
    def metrics(self) -> ICConnectionMetrics:
        """Return connection metrics."""
        return self._metrics

    @property
    def system_info(self) -> ICSystemInfo | None:
        """Return cached system information."""
        return self._system_info

    @property
    def connected(self) -> bool:
        """Return True if connected."""
        return self._connection is not None and self._connection.connected

    def set_disconnected_callback(
        self, callback: Callable[[ICBaseController, Exception | None], None] | None
    ) -> None:
        """Set callback for disconnection events."""
        self._disconnected_callback = callback

    async def start(self) -> None:
        """Connect and retrieve system information.

        Raises:
            ICConnectionError: If connection fails
            ICCommandError: If system info request fails
        """
        # Create connection
        self._connection = ICConnection(
            self._host,
            self._port,
            keepalive_interval=self._keepalive_interval,
        )

        # Set disconnect callback
        self._connection.set_disconnect_callback(self._on_disconnect)

        # Connect
        await self._connection.connect()
        self._metrics.successful_connects += 1

        _LOGGER.debug("Connected to IC at %s:%s", self._host, self._port)

        # Fetch system info
        with _RequestContext(self._metrics):
            try:
                response = await self._connection.send_request(
                    "GetParamList",
                    condition=f"{OBJTYP_ATTR}={SYSTEM_TYPE}",
                    objectList=[{"objnam": "INCR", "keys": ICSystemInfo.ATTRIBUTES_LIST}],
                )
                info = response["objectList"][0]
                self._system_info = ICSystemInfo(info["objnam"], info["params"])
            except ICResponseError as err:
                raise ICCommandError(err.code) from err

    async def stop(self) -> None:
        """Stop the controller and disconnect."""
        if self._connection:
            await self._connection.disconnect()
            self._connection = None

    def _on_disconnect(self, exc: Exception | None) -> None:
        """Handle disconnection from connection layer."""
        if self._disconnected_callback:
            self._disconnected_callback(self, exc)

    async def send_cmd(
        self,
        cmd: str,
        extra: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Send a command and return the response.

        Args:
            cmd: Command name (e.g., "GetParamList")
            extra: Additional parameters

        Returns:
            Response dictionary

        Raises:
            ICConnectionError: If not connected
            ICCommandError: If command fails
        """
        if not self._connection or not self._connection.connected:
            raise ICConnectionError("Not connected")

        with _RequestContext(self._metrics):
            try:
                return await self._connection.send_request(cmd, **(extra or {}))
            except ICResponseError as err:
                raise ICCommandError(err.code) from err

    async def request_changes(
        self,
        objnam: str,
        changes: dict[str, Any],
    ) -> dict[str, Any]:
        """Submit changes for an object.

        Args:
            objnam: Object name to modify
            changes: Attribute changes to apply

        Returns:
            Response dictionary
        """
        return await self.send_cmd(
            "SETPARAMLIST",
            {"objectList": [{"objnam": objnam, "params": changes}]},
        )

    async def get_all_objects(self, attribute_list: list[str]) -> list[ObjectEntry]:
        """Fetch attributes for all objects.

        Args:
            attribute_list: Attributes to fetch

        Returns:
            List of objects with their attributes
        """
        result = await self.send_cmd(
            "GetParamList",
            {"condition": "", "objectList": [{"objnam": "INCR", "keys": attribute_list}]},
        )
        pruned: list[ObjectEntry] = prune(result["objectList"])
        return pruned

    async def get_query(self, query_name: str, arguments: str = "") -> list[dict[str, Any]]:
        """Execute a query.

        Args:
            query_name: Query name
            arguments: Optional arguments

        Returns:
            Query results
        """
        result = await self.send_cmd("GetQuery", {"queryName": query_name, "arguments": arguments})
        answer: list[dict[str, Any]] = result["answer"]
        return answer

    async def get_configuration(self) -> list[dict[str, Any]]:
        """Get system configuration with bodies and circuits.

        This matches node-intellicenter's GetQuery with queryName="GetConfiguration".

        Returns:
            List of configuration objects including bodies and circuits
        """
        return await self.get_query("GetConfiguration")


class ICModelController(ICBaseController):
    """Controller that maintains a PoolModel of equipment state."""

    def __init__(
        self,
        host: str,
        model: PoolModel,
        port: int = 6681,
        keepalive_interval: float | None = None,
    ) -> None:
        """Initialize the controller.

        Args:
            host: IP address or hostname of IntelliCenter
            model: PoolModel to populate and update
            port: TCP port (default: 6681)
            keepalive_interval: Seconds between keepalive requests
        """
        super().__init__(host, port, keepalive_interval)
        self._model = model
        self._updated_callback: (
            Callable[[ICModelController, dict[str, dict[str, Any]]], None] | None
        ) = None

    def __repr__(self) -> str:
        return (
            f"ICModelController(host={self._host!r}, port={self._port}, "
            f"connected={self.connected}, objects={self._model.num_objects})"
        )

    @property
    def model(self) -> PoolModel:
        """Return the model."""
        return self._model

    def set_updated_callback(
        self, callback: Callable[[ICModelController, dict[str, dict[str, Any]]], None] | None
    ) -> None:
        """Set callback for model updates."""
        self._updated_callback = callback

    async def start(self) -> None:
        """Connect, fetch objects, and start monitoring.

        Raises:
            ICConnectionError: If connection fails
            ICCommandError: If initialization fails
        """
        await super().start()

        # Set notification callback
        if self._connection:
            self._connection.set_notification_callback(self._on_notification)

        # Fetch all objects
        all_objects = await self.get_all_objects(
            [OBJTYP_ATTR, SUBTYP_ATTR, SNAME_ATTR, PARENT_ATTR]
        )
        self._model.add_objects(all_objects)
        _LOGGER.info("Model contains %d objects", self._model.num_objects)

        # Request monitoring of attributes in batches
        attributes = self._model.attributes_to_track()
        query: list[dict[str, Any]] = []
        num_attributes = 0

        for items in attributes:
            query.append(items)
            num_attributes += len(items["keys"])

            # Batch to avoid overwhelming the system
            if num_attributes >= MAX_ATTRIBUTES_PER_QUERY:
                res = await self.send_cmd("RequestParamList", {"objectList": query})
                self._apply_updates(res["objectList"])
                query = []
                num_attributes = 0

        # Send remaining
        if query:
            res = await self.send_cmd("RequestParamList", {"objectList": query})
            self._apply_updates(res["objectList"])

    def _on_notification(self, msg: dict[str, Any]) -> None:
        """Handle NotifyList notifications."""
        if msg.get("command") == "NotifyList":
            try:
                self._apply_updates(msg["objectList"])
            except (KeyError, TypeError, ValueError) as err:
                _LOGGER.exception("Error processing NotifyList: %s", err)

    def _apply_updates(self, changes_as_list: list[ObjectEntry]) -> dict[str, dict[str, Any]]:
        """Apply updates to the model."""
        updates = self._model.process_updates(changes_as_list)

        # Update ICSystemInfo if changed
        if self._system_info and self._system_info.objnam in updates:
            self._system_info.update(updates[self._system_info.objnam])

        # Notify callback
        if updates and self._updated_callback:
            self._updated_callback(self, updates)

        return updates

    # --------------------------------------------------------------------------
    # Convenience methods for common operations
    # --------------------------------------------------------------------------

    async def set_circuit_state(self, objnam: str, state: bool) -> dict[str, Any]:
        """Set a circuit on or off.

        Args:
            objnam: Object name of the circuit
            state: True for ON, False for OFF

        Returns:
            Response dictionary
        """
        return await self.request_changes(objnam, {STATUS_ATTR: STATUS_ON if state else STATUS_OFF})

    async def set_multiple_circuit_states(self, objnams: list[str], state: bool) -> dict[str, Any]:
        """Set multiple circuits on or off simultaneously.

        This matches node-intellicenter's SetObjectStatus(array, boolean) functionality.

        Args:
            objnams: List of object names to control
            state: True for ON, False for OFF

        Returns:
            Response dictionary
        """
        status = STATUS_ON if state else STATUS_OFF
        object_list = [{"objnam": objnam, "params": {STATUS_ATTR: status}} for objnam in objnams]
        return await self.send_cmd("SETPARAMLIST", {"objectList": object_list})

    async def set_heat_mode(self, body_objnam: str, mode: HeaterType) -> dict[str, Any]:
        """Set the heat mode for a body of water.

        Args:
            body_objnam: Object name of the body (pool or spa)
            mode: HeaterType enum value

        Returns:
            Response dictionary

        Example:
            await controller.set_heat_mode("B1101", HeaterType.HEATER)
        """
        return await self.request_changes(body_objnam, {MODE_ATTR: str(mode.value)})

    async def set_setpoint(self, body_objnam: str, temperature: int) -> dict[str, Any]:
        """Set the temperature setpoint for a body of water.

        Args:
            body_objnam: Object name of the body (pool or spa)
            temperature: Target temperature (units match system config)

        Returns:
            Response dictionary
        """
        return await self.request_changes(body_objnam, {LOTMP_ATTR: str(temperature)})

    async def set_super_chlorinate(self, chem_objnam: str, enabled: bool) -> dict[str, Any]:
        """Enable or disable super chlorination (boost mode).

        Args:
            chem_objnam: Object name of the chemistry controller
            enabled: True to enable, False to disable

        Returns:
            Response dictionary
        """
        return await self.request_changes(
            chem_objnam, {SUPER_ATTR: STATUS_ON if enabled else STATUS_OFF}
        )

    def get_bodies(self) -> list[Any]:
        """Get all body objects (pools and spas).

        Returns:
            List of PoolObject for bodies
        """
        return [obj for obj in self._model if obj.objtype == BODY_TYPE]

    def get_circuits(self) -> list[Any]:
        """Get all circuit objects.

        Returns:
            List of PoolObject for circuits
        """
        return [obj for obj in self._model if obj.objtype == CIRCUIT_TYPE]

    def get_heaters(self) -> list[Any]:
        """Get all heater objects.

        Returns:
            List of PoolObject for heaters
        """
        return [obj for obj in self._model if obj.objtype == HEATER_TYPE]

    def get_schedules(self) -> list[Any]:
        """Get all schedule objects.

        Returns:
            List of PoolObject for schedules
        """
        return [obj for obj in self._model if obj.objtype == SCHED_TYPE]

    def get_sensors(self) -> list[Any]:
        """Get all sensor objects.

        Returns:
            List of PoolObject for sensors
        """
        return [obj for obj in self._model if obj.objtype == SENSE_TYPE]

    def get_pumps(self) -> list[Any]:
        """Get all pump objects.

        Returns:
            List of PoolObject for pumps
        """
        return [obj for obj in self._model if obj.objtype == PUMP_TYPE]

    def get_chem_controllers(self) -> list[Any]:
        """Get all chemistry controller objects (IntelliChem, IntelliChlor).

        Returns:
            List of PoolObject for chemistry controllers
        """
        return [obj for obj in self._model if obj.objtype == CHEM_TYPE]


# Reconnection constants
DEFAULT_RECONNECT_DELAY = 30
DEFAULT_DISCONNECT_DEBOUNCE = 15
MAX_RECONNECT_DELAY = 600
CIRCUIT_BREAKER_FAILURES = 5
CIRCUIT_BREAKER_RESET_TIME = 300


@runtime_checkable
class ICConnectionHandlerCallbacks(Protocol):
    """Protocol for ICConnectionHandler event callbacks.

    Implement this protocol to handle connection lifecycle events.
    """

    def on_started(self, controller: ICBaseController) -> None:
        """Called on initial successful connection."""
        ...

    def on_reconnected(self, controller: ICBaseController) -> None:
        """Called when reconnected after a disconnect."""
        ...

    def on_disconnected(self, controller: ICBaseController, exc: Exception | None) -> None:
        """Called when disconnected (after debounce period)."""
        ...

    def on_retrying(self, delay: int) -> None:
        """Called before each retry attempt."""
        ...

    def on_updated(self, controller: ICModelController, updates: dict[str, dict[str, Any]]) -> None:
        """Called when model is updated (only for ICModelController)."""
        ...


class ICConnectionHandler:
    """Manages automatic reconnection with exponential backoff.

    This handler wraps a controller and provides automatic reconnection
    with exponential backoff, circuit breaker pattern, and debounced
    disconnect notifications.

    Example:
        model = PoolModel()
        controller = ICModelController("192.168.1.100", model)
        handler = ICConnectionHandler(controller)

        # Override callbacks
        handler.on_started = lambda ctrl: print("Connected!")
        handler.on_disconnected = lambda ctrl, exc: print(f"Disconnected: {exc}")

        await handler.start()
    """

    def __init__(
        self,
        controller: ICBaseController,
        time_between_reconnects: int = DEFAULT_RECONNECT_DELAY,
        disconnect_debounce_time: int = DEFAULT_DISCONNECT_DEBOUNCE,
    ) -> None:
        """Initialize the handler.

        Args:
            controller: Controller to manage
            time_between_reconnects: Initial reconnect delay (seconds)
            disconnect_debounce_time: Grace period before disconnect notification
        """
        self._controller = controller
        self._time_between_reconnects = time_between_reconnects
        self._disconnect_debounce_time = disconnect_debounce_time

        self._starter_task: asyncio.Task[None] | None = None
        self._disconnect_debounce_task: asyncio.Task[None] | None = None
        self._stopped = False
        self._first_time = True
        self._is_connected = False

        # Circuit breaker
        self._failure_count = 0
        self._last_failure_time: float | None = None

        # Set callbacks on controller
        controller.set_disconnected_callback(self._on_disconnect)

        if isinstance(controller, ICModelController):
            controller.set_updated_callback(self._on_model_updated)

    def __repr__(self) -> str:
        return (
            f"ICConnectionHandler(controller={self._controller!r}, "
            f"connected={self._is_connected}, failures={self._failure_count})"
        )

    @property
    def controller(self) -> ICBaseController:
        """Return the managed controller."""
        return self._controller

    async def start(self) -> None:
        """Start the connection handler."""
        if not self._starter_task:
            self._starter_task = asyncio.create_task(self._starter())

    def stop(self) -> None:
        """Stop the handler and controller."""
        self._stopped = True
        if self._starter_task:
            self._starter_task.cancel()
            self._starter_task = None
        if self._disconnect_debounce_task:
            self._disconnect_debounce_task.cancel()
            self._disconnect_debounce_task = None
        # Fire and forget the async stop
        asyncio.create_task(self._stop_controller())

    async def _stop_controller(self) -> None:
        """Stop the controller asynchronously."""
        with contextlib.suppress(Exception):
            await self._controller.stop()

    async def _starter(self, initial_delay: int = 0) -> None:
        """Attempt to connect with exponential backoff."""
        delay = self._time_between_reconnects

        while not self._stopped:
            try:
                # Check circuit breaker reset
                if (
                    self._last_failure_time
                    and time.monotonic() - self._last_failure_time > CIRCUIT_BREAKER_RESET_TIME
                ):
                    self._failure_count = 0
                    self._last_failure_time = None

                # Circuit breaker open - pause
                if self._failure_count >= CIRCUIT_BREAKER_FAILURES:
                    _LOGGER.warning(
                        "Circuit breaker open - pausing %ds", CIRCUIT_BREAKER_RESET_TIME
                    )
                    await asyncio.sleep(CIRCUIT_BREAKER_RESET_TIME)
                    self._failure_count = 0

                if initial_delay:
                    self.on_retrying(initial_delay)
                    self._controller._metrics.reconnect_attempts += 1
                    await asyncio.sleep(initial_delay)
                    initial_delay = 0

                await self._controller.start()

                # Success - reset circuit breaker
                self._failure_count = 0
                self._last_failure_time = None

                if self._disconnect_debounce_task:
                    self._disconnect_debounce_task.cancel()
                    self._disconnect_debounce_task = None

                if self._first_time:
                    self.on_started(self._controller)
                    self._first_time = False
                elif not self._is_connected:
                    self.on_reconnected(self._controller)

                self._is_connected = True
                self._starter_task = None
                return

            except (TimeoutError, OSError, ICConnectionError, ICCommandError) as err:
                self._failure_count += 1
                self._last_failure_time = time.monotonic()
                self._controller._metrics.reconnect_attempts += 1

                _LOGGER.error(
                    "Connection failed: %s (failure %d/%d)",
                    err,
                    self._failure_count,
                    CIRCUIT_BREAKER_FAILURES,
                )
                self.on_retrying(delay)
                await asyncio.sleep(delay)
                delay = min(int(delay * 1.5), MAX_RECONNECT_DELAY)

    def _on_disconnect(self, controller: ICBaseController, exc: Exception | None) -> None:
        """Handle disconnection."""
        if self._stopped:
            return

        _LOGGER.warning("Disconnected from %s: %s", controller.host, exc)
        self._is_connected = False

        # Debounced disconnect notification
        if self._disconnect_debounce_task:
            self._disconnect_debounce_task.cancel()
        self._disconnect_debounce_task = asyncio.create_task(
            self._delayed_disconnect(controller, exc)
        )

        # Start reconnection
        self._starter_task = asyncio.create_task(self._starter(self._time_between_reconnects))

    async def _delayed_disconnect(
        self, controller: ICBaseController, exc: Exception | None
    ) -> None:
        """Notify about disconnection after debounce period."""
        try:
            await asyncio.sleep(self._disconnect_debounce_time)
            if not self._is_connected:
                self.on_disconnected(controller, exc)
        except asyncio.CancelledError:
            pass

    def _on_model_updated(
        self, controller: ICModelController, updates: dict[str, dict[str, Any]]
    ) -> None:
        """Internal callback that forwards to user callback."""
        self.on_updated(controller, updates)

    # Override these methods or assign callables to handle events
    def on_started(self, controller: ICBaseController) -> None:
        """Called on initial connection. Override or replace to handle."""

    def on_reconnected(self, controller: ICBaseController) -> None:
        """Called on reconnection after disconnect. Override or replace to handle."""

    def on_disconnected(self, controller: ICBaseController, exc: Exception | None) -> None:
        """Called after debounce period if still disconnected. Override or replace to handle."""

    def on_retrying(self, delay: int) -> None:
        """Called before retry attempt. Override or replace to handle."""
        _LOGGER.info("Retrying in %ds", delay)

    def on_updated(self, controller: ICModelController, updates: dict[str, dict[str, Any]]) -> None:
        """Called when model is updated. Override or replace to handle."""



"""Body of water (pool/spa) attribute definitions."""

from .constants import (
    BOOST_ATTR,
    CIRCUIT_ATTR,
    HEATER_ATTR,
    HNAME_ATTR,
    HTMODE_ATTR,
    LISTORD_ATTR,
    LOTMP_ATTR,
    LSTTMP_ATTR,
    MODE_ATTR,
    PARENT_ATTR,
    PRIM_ATTR,
    READY_ATTR,
    SEC_ATTR,
    SELECT_ATTR,
    SNAME_ATTR,
    SPEED_ATTR,
    STATIC_ATTR,
    STATUS_ATTR,
    SUBTYP_ATTR,
    TEMP_ATTR,
    VOL_ATTR,
)

# Represents a body of water (pool or spa)
# Matches node-intellicenter GetBodyStatus attributes
BODY_ATTRIBUTES = {
    "ACT1",  # (int) Activity setting 1
    "ACT2",  # (int) Activity setting 2
    "ACT3",  # (int) Activity setting 3
    "ACT4",  # (int) Activity setting 4
    BOOST_ATTR,  # (ON/OFF) Boost heating enabled
    CIRCUIT_ATTR,  # (objnam) Associated circuit
    "FILTER",  # (objnam) Circuit object that filters this body
    HEATER_ATTR,  # (objnam) Associated heater
    "HITMP",  # (int) Maximum temperature setting
    HNAME_ATTR,  # equals to OBJNAM
    HTMODE_ATTR,  # (int) >0 if currently heating, 0 if not
    "HTSRC",  # (objnam) the heating source (or '00000')
    LISTORD_ATTR,  # (int) used to order in UI
    LOTMP_ATTR,  # (int) Desired temperature setpoint
    LSTTMP_ATTR,  # (int) Last recorded temperature
    "MANHT",  # Manual heating
    "MANUAL",  # (int) Manual mode
    MODE_ATTR,  # (str) Current mode
    PARENT_ATTR,  # (objnam) parent object
    PRIM_ATTR,  # (int) Primary setting
    READY_ATTR,  # (ON/OFF) Ready state
    SEC_ATTR,  # (int) Secondary setting
    SELECT_ATTR,  # (str) Selection mode
    "SETPT",  # (int) Set point (same as LOTMP)
    "SHARE",  # (objnam) Sharing with other body
    SNAME_ATTR,  # (str) Friendly name
    SPEED_ATTR,  # (int) Speed setting
    "SRCTYP",  # Source type (e.g., "GENERIC")
    STATIC_ATTR,  # (ON/OFF) Static setting
    STATUS_ATTR,  # (ON/OFF) 'ON' if body is "active"
    SUBTYP_ATTR,  # 'POOL' or 'SPA'
    TEMP_ATTR,  # (int) Current temperature
    VOL_ATTR,  # (int) Volume in Gallons
}



"""Miscellaneous attribute definitions (remote, valve, external instruments, etc.)."""

from .constants import (
    BODY_ATTR,
    CIRCUIT_ATTR,
    COMUART_ATTR,
    DLY_ATTR,
    ENABLE_ATTR,
    HNAME_ATTR,
    LISTORD_ATTR,
    NORMAL_ATTR,
    PARENT_ATTR,
    READY_ATTR,
    SHOMNU_ATTR,
    SNAME_ATTR,
    SOURCE_ATTR,
    STATIC_ATTR,
    STATUS_ATTR,
    SUBTYP_ATTR,
)

# External instrument attributes (covers, etc.)
EXTINSTR_ATTRIBUTES = {
    BODY_ATTR,  # (objnam) which body it covers
    HNAME_ATTR,  # equals to OBJNAM
    LISTORD_ATTR,  # (int) used to order in UI
    NORMAL_ATTR,  # (ON/OFF) 'ON' for Cover State Normally On
    PARENT_ATTR,  # (objnam)
    READY_ATTR,  # (ON/OFF) ???
    SNAME_ATTR,  # (str) friendly name
    STATIC_ATTR,  # (ON/OFF) 'OFF'
    STATUS_ATTR,  # (ON/OFF) 'ON' if cover enabled
    SUBTYP_ATTR,  # only seen 'COVER'
}

# Feature attributes (no idea what this represents)
FEATR_ATTRIBUTES = {
    HNAME_ATTR,
    LISTORD_ATTR,
    READY_ATTR,
    SNAME_ATTR,
    SOURCE_ATTR,
    STATIC_ATTR,
}

# Press attributes (no idea what this object type represents)
# Only seems to be one instance of it
PRESS_ATTRIBUTES = {
    SHOMNU_ATTR,  # (ON/OFF) ???
    SNAME_ATTR,  # seems equal to objnam
    STATIC_ATTR,  # (ON/OFF) only seen ON
}

# Remote button mapping attributes
REMBTN_ATTRIBUTES = {
    CIRCUIT_ATTR,  # (objnam) the circuit triggered by the button
    LISTORD_ATTR,  # (int) which button on the remote (1 to 4)
    PARENT_ATTR,  # (objnam) the remote this button is associated with
    STATIC_ATTR,  # (ON/OFF) not sure, only seen 'ON'
}

# Remote attributes
REMOTE_ATTRIBUTES = {
    BODY_ATTR,  # (objnam) the body the remote controls
    COMUART_ATTR,  # X25 address?
    ENABLE_ATTR,  # (ON/OFF) 'ON' if the remote is set to active
    HNAME_ATTR,  # same as objnam
    LISTORD_ATTR,  # number likely used to order things in UI
    SNAME_ATTR,  # friendly name
    STATIC_ATTR,  # (ON/OFF) not sure, only seen 'OFF'
    SUBTYP_ATTR,  # type of the remote, I've only seen IS4
}

# Valve attributes
VALVE_ATTRIBUTES = {
    "ASSIGN",  # 'NONE', 'INTAKE' or 'RETURN'
    CIRCUIT_ATTR,  # I've only seen '00000'
    DLY_ATTR,  # (ON/OFF)
    HNAME_ATTR,  # same as objnam
    PARENT_ATTR,  # (objnam) parent (a module)
    SNAME_ATTR,  # friendly name
    STATIC_ATTR,  # (ON/OFF) I've only seen 'OFF'
    SUBTYP_ATTR,  # I've only seen 'LEGACY'
}



"""Exception hierarchy for pyintellicenter.

This module provides a unified exception hierarchy for all library errors,
making it easier to catch and handle specific error types.
"""

from __future__ import annotations


class ICError(Exception):
    """Base exception for all pyintellicenter errors.

    All exceptions raised by this library inherit from this class,
    making it easy to catch any library-specific error.
    """


class ICConnectionError(ICError):
    """Raised when connection fails or is lost.

    This exception is raised for network-level issues such as:
    - Failed to establish TCP connection
    - Connection timeout
    - Connection closed unexpectedly
    - Keepalive timeout
    """


class ICResponseError(ICError):
    """Raised when IntelliCenter returns an error response.

    The IntelliCenter system returns HTTP-like status codes.
    A response code other than "200" indicates an error.

    Attributes:
        code: The error code returned by IntelliCenter
        message: Optional error message
    """

    def __init__(self, code: str, message: str | None = None) -> None:
        self.code = code
        self.message = message
        super().__init__(
            f"IntelliCenter error {code}: {message}" if message else f"IntelliCenter error {code}"
        )

    def __repr__(self) -> str:
        return f"ICResponseError(code={self.code!r}, message={self.message!r})"


class ICCommandError(ICError):
    """Raised when a command to IntelliCenter fails.

    This is a higher-level exception that wraps ICResponseError
    for use at the controller layer.

    Attributes:
        error_code: The error code from the underlying response
    """

    def __init__(self, error_code: str) -> None:
        self._error_code = error_code
        super().__init__(f"IntelliCenter command error: {error_code}")

    @property
    def error_code(self) -> str:
        """Return the error code."""
        return self._error_code

    def __repr__(self) -> str:
        return f"ICCommandError(error_code={self._error_code!r})"


class ICTimeoutError(ICError):
    """Raised when a request times out waiting for a response.

    This is distinct from connection timeout - it indicates the
    connection is established but the response was not received
    within the expected time.
    """



"""pyintellicenter - Python library for Pentair IntelliCenter pool control systems.

This library provides the core protocol and model classes for communicating
with Pentair IntelliCenter pool control systems over local network.

Example usage:
    ```python
    import asyncio
    from pyintellicenter import ICModelController, PoolModel, ICConnectionHandler

    async def main():
        model = PoolModel()
        controller = ICModelController("192.168.1.100", model)
        handler = ICConnectionHandler(controller)
        await handler.start()

        # Access pool equipment
        for obj in model:
            print(f"{obj.sname}: {obj.status}")

    asyncio.run(main())
    ```
"""

# Re-export all public names from submodules
from .attributes import (
    # Attribute name constants
    ACT_ATTR,
    AVAIL_ATTR,
    BODY_ATTR,
    # Type constants
    BODY_TYPE,
    BOOST_ATTR,
    CHEM_TYPE,
    CIRCGRP_TYPE,
    CIRCUIT_ATTR,
    CIRCUIT_TYPE,
    # Status constants
    COLOR_EFFECT_SUBTYPES,
    COMUART_ATTR,
    COOLING_ATTR,
    DLY_ATTR,
    ENABLE_ATTR,
    EXTINSTR_TYPE,
    FEATR_ATTR,
    FEATR_TYPE,
    GPM_ATTR,
    GROUP_ATTR,
    HEATER_ATTR,
    HEATER_TYPE,
    HEATING_ATTR,
    HNAME_ATTR,
    HTMODE_ATTR,
    LIGHT_EFFECTS,
    LIGHT_SUBTYPES,
    LISTORD_ATTR,
    LOTMP_ATTR,
    LSTTMP_ATTR,
    MODE_ATTR,
    MODULE_TYPE,
    NORMAL_ATTR,
    # Special values
    NULL_OBJNAM,
    OBJTYP_ATTR,
    ORPTNK_ATTR,
    ORPVAL_ATTR,
    PANEL_TYPE,
    PARENT_ATTR,
    PERMIT_ATTR,
    PERMIT_TYPE,
    PHTNK_ATTR,
    PHVAL_ATTR,
    PMPCIRC_TYPE,
    PRESS_TYPE,
    PRIM_ATTR,
    PROPNAME_ATTR,
    PUMP_STATUS_OFF,
    PUMP_STATUS_ON,
    PUMP_TYPE,
    PWR_ATTR,
    QUALTY_ATTR,
    READY_ATTR,
    REMBTN_TYPE,
    REMOTE_TYPE,
    RPM_ATTR,
    SALT_ATTR,
    SCHED_TYPE,
    SEC_ATTR,
    SELECT_ATTR,
    SENSE_TYPE,
    SHOMNU_ATTR,
    SMTSRT_ATTR,
    SNAME_ATTR,
    SOURCE_ATTR,
    SPEED_ATTR,
    STATIC_ATTR,
    STATUS_ATTR,
    STATUS_OFF,
    STATUS_ON,
    SUBTYP_ATTR,
    SUPER_ATTR,
    SYSTEM_TYPE,
    SYSTIM_TYPE,
    TEMP_ATTR,
    TIME_ATTR,
    TIMOUT_ATTR,
    UPDATE_ATTR,
    USE_ATTR,
    USER_PRIVILEGES,
    VACFLO_ATTR,
    VACTIM_ATTR,
    VALVE_TYPE,
    VER_ATTR,
    VOL_ATTR,
    # Enums
    HeaterType,
)
from .connection import ICConnection
from .controller import (
    ICBaseController,
    ICConnectionHandler,
    ICConnectionHandlerCallbacks,
    ICConnectionMetrics,
    ICModelController,
    ICSystemInfo,
)
from .exceptions import (
    ICCommandError,
    ICConnectionError,
    ICError,
    ICResponseError,
    ICTimeoutError,
)
from .model import PoolModel, PoolObject
from .types import (
    NotificationMessage,
    ObjectEntry,
    ObjectListRequest,
    ObjectParams,
    ResponseMessage,
)

# Discovery module (requires optional 'zeroconf' dependency)
# Import conditionally to avoid ImportError when zeroconf is not installed
try:
    from .discovery import (  # noqa: F401
        DEFAULT_DISCOVERY_TIMEOUT,
        ICUnit,
        discover_intellicenter_units,
        find_unit_by_host,
        find_unit_by_name,
    )

    _DISCOVERY_AVAILABLE = True
except ImportError:
    _DISCOVERY_AVAILABLE = False

__version__ = "0.0.5a3"

__all__ = [
    # Version
    "__version__",
    # Exceptions
    "ICError",
    "ICConnectionError",
    "ICResponseError",
    "ICCommandError",
    "ICTimeoutError",
    # Connection
    "ICConnection",
    # Controller classes
    "ICBaseController",
    "ICModelController",
    "ICConnectionHandler",
    "ICConnectionHandlerCallbacks",
    "ICConnectionMetrics",
    "ICSystemInfo",
    # Model classes
    "PoolModel",
    "PoolObject",
    # Type definitions
    "NotificationMessage",
    "ObjectEntry",
    "ObjectListRequest",
    "ObjectParams",
    "ResponseMessage",
    # Enums
    "HeaterType",
    # Status constants
    "STATUS_ON",
    "STATUS_OFF",
    "PUMP_STATUS_ON",
    "PUMP_STATUS_OFF",
    "LIGHT_SUBTYPES",
    "COLOR_EFFECT_SUBTYPES",
    "LIGHT_EFFECTS",
    # Object types
    "BODY_TYPE",
    "CHEM_TYPE",
    "CIRCUIT_TYPE",
    "CIRCGRP_TYPE",
    "EXTINSTR_TYPE",
    "FEATR_TYPE",
    "HEATER_TYPE",
    "MODULE_TYPE",
    "PANEL_TYPE",
    "PERMIT_TYPE",
    "PMPCIRC_TYPE",
    "PRESS_TYPE",
    "PUMP_TYPE",
    "REMBTN_TYPE",
    "REMOTE_TYPE",
    "SCHED_TYPE",
    "SENSE_TYPE",
    "SYSTEM_TYPE",
    "SYSTIM_TYPE",
    "VALVE_TYPE",
    # Special values
    "NULL_OBJNAM",
    # Attributes
    "ACT_ATTR",
    "AVAIL_ATTR",
    "BODY_ATTR",
    "BOOST_ATTR",
    "CIRCUIT_ATTR",
    "COMUART_ATTR",
    "COOLING_ATTR",
    "DLY_ATTR",
    "ENABLE_ATTR",
    "FEATR_ATTR",
    "GPM_ATTR",
    "GROUP_ATTR",
    "HEATER_ATTR",
    "HEATING_ATTR",
    "HNAME_ATTR",
    "HTMODE_ATTR",
    "LISTORD_ATTR",
    "LOTMP_ATTR",
    "LSTTMP_ATTR",
    "MODE_ATTR",
    "NORMAL_ATTR",
    "OBJTYP_ATTR",
    "ORPTNK_ATTR",
    "ORPVAL_ATTR",
    "PARENT_ATTR",
    "PERMIT_ATTR",
    "PHTNK_ATTR",
    "PHVAL_ATTR",
    "PRIM_ATTR",
    "PROPNAME_ATTR",
    "PWR_ATTR",
    "QUALTY_ATTR",
    "READY_ATTR",
    "RPM_ATTR",
    "SALT_ATTR",
    "SEC_ATTR",
    "SELECT_ATTR",
    "SHOMNU_ATTR",
    "SMTSRT_ATTR",
    "SNAME_ATTR",
    "SOURCE_ATTR",
    "SPEED_ATTR",
    "STATIC_ATTR",
    "STATUS_ATTR",
    "SUBTYP_ATTR",
    "SUPER_ATTR",
    "TEMP_ATTR",
    "TIME_ATTR",
    "TIMOUT_ATTR",
    "UPDATE_ATTR",
    "USE_ATTR",
    "USER_PRIVILEGES",
    "VACFLO_ATTR",
    "VACTIM_ATTR",
    "VER_ATTR",
    "VOL_ATTR",
]

# Add discovery exports if available
if _DISCOVERY_AVAILABLE:
    __all__.extend(
        [
            "ICUnit",
            "discover_intellicenter_units",
            "find_unit_by_name",
            "find_unit_by_host",
            "DEFAULT_DISCOVERY_TIMEOUT",
        ]
    )



"""Circuit and circuit group attribute definitions."""

from .constants import (
    ACT_ATTR,
    BODY_ATTR,
    CIRCUIT_ATTR,
    DLY_ATTR,
    FEATR_ATTR,
    HNAME_ATTR,
    LISTORD_ATTR,
    PARENT_ATTR,
    READY_ATTR,
    SELECT_ATTR,
    SHOMNU_ATTR,
    SNAME_ATTR,
    STATIC_ATTR,
    STATUS_ATTR,
    SUBTYP_ATTR,
    TIME_ATTR,
    USE_ATTR,
)

# Circuit attributes
CIRCUIT_ATTRIBUTES = {
    ACT_ATTR,  # to be set for changing USE attribute
    BODY_ATTR,
    "CHILD",
    "COVER",
    "DNTSTP",  # (ON/OFF) "Don't Stop", disable egg timer
    FEATR_ATTR,  # (ON/OFF) Featured
    "FREEZE",  # (ON/OFF) Freeze Protection
    HNAME_ATTR,  # equals to OBJNAM
    "LIMIT",
    LISTORD_ATTR,  # (int) used to order in UI
    "OBJLIST",
    PARENT_ATTR,  # OBJNAM of the parent object
    READY_ATTR,  # (ON/OFF) ??
    SELECT_ATTR,  # ???
    "SET",  # (ON/OFF) for light groups only
    SHOMNU_ATTR,  # (str) permissions
    SNAME_ATTR,  # (str) friendly name
    STATIC_ATTR,  # (ON/OFF) ??
    STATUS_ATTR,  # (ON/OFF) 'ON' if circuit is active
    SUBTYP_ATTR,  # subtype can be '?
    "SWIM",  # (ON/OFF) for light groups only
    "SYNC",  # (ON/OFF) for light groups only
    TIME_ATTR,  # (int) Egg Timer, number of minutes
    "USAGE",
    USE_ATTR,  # for lights with light effects, indicate the 'color'
}

# Circuit group attributes
CIRCGRP_ATTRIBUTES = {
    ACT_ATTR,
    CIRCUIT_ATTR,
    DLY_ATTR,
    LISTORD_ATTR,
    PARENT_ATTR,
    READY_ATTR,
    STATIC_ATTR,
}



"""Type definitions for pyintellicenter.

This module provides TypedDict definitions for structured data
used throughout the library, improving type safety and IDE support.
"""

from __future__ import annotations

from typing import Any, TypedDict


class ObjectParams(TypedDict, total=False):
    """Parameters for a pool object.

    This is a partial dict since not all attributes are always present.
    """

    OBJTYP: str
    SUBTYP: str
    SNAME: str
    PARENT: str
    STATUS: str


class ObjectEntry(TypedDict):
    """An entry in an objectList response."""

    objnam: str
    params: dict[str, Any]


class ObjectListRequest(TypedDict, total=False):
    """Request structure for objectList queries."""

    objnam: str
    keys: list[str]


class NotificationMessage(TypedDict):
    """Structure of a NotifyList notification message."""

    command: str
    objectList: list[ObjectEntry]


class ResponseMessage(TypedDict, total=False):
    """Structure of a response message from IntelliCenter."""

    response: str
    command: str
    messageID: str
    objectList: list[ObjectEntry]
    answer: list[dict[str, Any]]



"""Modern asyncio connection for Pentair IntelliCenter.

This module provides a clean, modern asyncio streams-based connection
to Pentair IntelliCenter pool control systems. It replaces the legacy
Protocol-based implementation with simpler async/await patterns.

Features:
- asyncio.open_connection() for TCP streams
- Automatic message framing via readline() (messages end with \\r\\n)
- asyncio.timeout() for clean timeout handling
- asyncio.Lock for flow control (one request at a time)
- Automatic keepalive with configurable interval
- Support for both sync and async notification callbacks
"""

from __future__ import annotations

import asyncio
import contextlib
import inspect
import logging
from typing import TYPE_CHECKING, Any

import orjson

from .exceptions import ICConnectionError, ICResponseError

if TYPE_CHECKING:
    from asyncio import StreamReader, StreamWriter
    from collections.abc import Awaitable, Callable

    # Callback types
    NotificationCallback = Callable[[dict[str, Any]], None | Awaitable[None]]
    DisconnectCallback = Callable[[Exception | None], None]

_LOGGER = logging.getLogger(__name__)

# Connection configuration
DEFAULT_PORT = 6681
RESPONSE_TIMEOUT = 30.0  # seconds to wait for a response
KEEPALIVE_INTERVAL = 90.0  # seconds between keepalive requests
CONNECTION_TIMEOUT = 10.0  # seconds to wait for initial connection


class ICConnection:
    """Modern asyncio connection to IntelliCenter.

    This class manages the TCP connection and provides a simple async interface
    for sending requests and receiving responses/notifications.

    Example:
        async with ICConnection("192.168.1.100") as conn:
            response = await conn.send_request(
                "GetParamList",
                objectList=[{"objnam": "INCR", "keys": ["VER"]}]
            )
            print(response)
    """

    def __init__(
        self,
        host: str,
        port: int = DEFAULT_PORT,
        response_timeout: float = RESPONSE_TIMEOUT,
        keepalive_interval: float = KEEPALIVE_INTERVAL,
    ) -> None:
        """Initialize connection configuration.

        Args:
            host: IP address or hostname of IntelliCenter
            port: TCP port (default: 6681)
            response_timeout: Seconds to wait for response (default: 30)
            keepalive_interval: Seconds between keepalive requests (default: 90)
        """
        self._host = host
        self._port = port
        self._response_timeout = response_timeout
        self._keepalive_interval = keepalive_interval

        # Connection state
        self._reader: StreamReader | None = None
        self._writer: StreamWriter | None = None
        self._connected = False

        # Flow control: one request at a time
        self._request_lock = asyncio.Lock()

        # Message ID counter
        self._message_id = 0

        # Callbacks (support both sync and async)
        self._notification_callback: NotificationCallback | None = None
        self._disconnect_callback: DisconnectCallback | None = None

        # Keepalive task
        self._keepalive_task: asyncio.Task[None] | None = None

    def __repr__(self) -> str:
        """Return a detailed string representation for debugging."""
        return f"ICConnection(host={self._host!r}, port={self._port}, connected={self.connected})"

    @property
    def host(self) -> str:
        """Return the host address."""
        return self._host

    @property
    def port(self) -> int:
        """Return the port number."""
        return self._port

    @property
    def connected(self) -> bool:
        """Return True if connected."""
        return self._connected and self._writer is not None and not self._writer.is_closing()

    def set_notification_callback(self, callback: NotificationCallback | None) -> None:
        """Set callback for NotifyList push notifications.

        The callback can be either sync or async. If async, it will be awaited.

        Args:
            callback: Function to call with notification data, or None to clear.
        """
        self._notification_callback = callback

    def set_disconnect_callback(self, callback: DisconnectCallback | None) -> None:
        """Set callback for disconnection events.

        Args:
            callback: Function to call on disconnect, or None to clear.
        """
        self._disconnect_callback = callback

    async def connect(self) -> None:
        """Establish connection to IntelliCenter.

        Raises:
            ICConnectionError: If connection fails or times out.
        """
        if self.connected:
            return

        try:
            async with asyncio.timeout(CONNECTION_TIMEOUT):
                self._reader, self._writer = await asyncio.open_connection(self._host, self._port)
            self._connected = True
            self._message_id = 0
            _LOGGER.debug("Connected to IC at %s:%s", self._host, self._port)

            # Start keepalive task
            self._keepalive_task = asyncio.create_task(self._keepalive_loop())

        except TimeoutError as err:
            raise ICConnectionError(f"Connection to {self._host}:{self._port} timed out") from err
        except OSError as err:
            raise ICConnectionError(
                f"Failed to connect to {self._host}:{self._port}: {err}"
            ) from err

    async def disconnect(self) -> None:
        """Close the connection gracefully."""
        self._connected = False

        # Cancel keepalive task
        if self._keepalive_task and not self._keepalive_task.done():
            self._keepalive_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._keepalive_task
            self._keepalive_task = None

        # Close writer
        if self._writer:
            self._writer.close()
            with contextlib.suppress(Exception):
                await self._writer.wait_closed()
            self._writer = None
            self._reader = None

        _LOGGER.debug("Disconnected from IC")

    async def __aenter__(self) -> ICConnection:
        """Async context manager entry."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Async context manager exit."""
        await self.disconnect()

    def _next_message_id(self) -> str:
        """Generate the next message ID."""
        self._message_id += 1
        return str(self._message_id)

    async def _read_message(self) -> dict[str, Any]:
        """Read a single JSON message from the connection.

        Returns:
            Parsed JSON message as a dictionary.

        Raises:
            ICConnectionError: If connection is closed or read fails.
        """
        if not self._reader:
            raise ICConnectionError("Not connected")

        try:
            line = await self._reader.readline()
            if not line:
                raise ICConnectionError("Connection closed by remote")

            result: dict[str, Any] = orjson.loads(line)
            return result

        except orjson.JSONDecodeError as err:
            _LOGGER.error("Invalid JSON received: %s", err)
            raise ICConnectionError(f"Invalid JSON: {err}") from err

    async def _write_message(self, message: dict[str, Any]) -> None:
        """Write a JSON message to the connection.

        Args:
            message: Dictionary to send as JSON.

        Raises:
            ICConnectionError: If connection is closed or write fails.
        """
        if not self._writer:
            raise ICConnectionError("Not connected")

        try:
            packet = orjson.dumps(message) + b"\r\n"
            self._writer.write(packet)
            await self._writer.drain()
        except OSError as err:
            raise ICConnectionError(f"Write failed: {err}") from err

    async def _invoke_notification_callback(self, msg: dict[str, Any]) -> None:
        """Invoke the notification callback, handling both sync and async callbacks."""
        if not self._notification_callback:
            return

        if inspect.iscoroutinefunction(self._notification_callback):
            await self._notification_callback(msg)
        else:
            self._notification_callback(msg)

    async def send_request(
        self,
        command: str,
        request_timeout: float | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Send a request and wait for the response.

        This method handles flow control (one request at a time) and
        processes any notifications received while waiting for the response.

        Args:
            command: The command name (e.g., "GetParamList", "SetParamList")
            request_timeout: Override response timeout (default: use instance timeout)
            **kwargs: Additional fields to include in the request

        Returns:
            The response message dictionary.

        Raises:
            ICConnectionError: If not connected or connection fails.
            ICResponseError: If IntelliCenter returns an error response.
            TimeoutError: If no response received within timeout.
        """
        if not self.connected:
            raise ICConnectionError("Not connected")

        # Build request with message ID
        request: dict[str, Any] = {
            "messageID": self._next_message_id(),
            "command": command,
            **kwargs,
        }

        effective_timeout = (
            request_timeout if request_timeout is not None else self._response_timeout
        )

        # Flow control: one request at a time
        async with self._request_lock:
            await self._write_message(request)
            _LOGGER.debug("Sent request: %s (ID: %s)", command, request["messageID"])

            # Read until we get a response (not a notification)
            try:
                async with asyncio.timeout(effective_timeout):
                    while True:
                        msg = await self._read_message()

                        # Check if this is a response (has "response" field)
                        if "response" in msg:
                            response_code: str = msg.get("response", "unknown")
                            if response_code != "200":
                                raise ICResponseError(response_code)
                            _LOGGER.debug("Received response for %s", msg.get("command"))
                            return dict(msg)

                        # It's a notification (NotifyList) - process and continue
                        if msg.get("command") == "NotifyList":
                            _LOGGER.debug("Received NotifyList notification")
                            await self._invoke_notification_callback(msg)

            except TimeoutError:
                _LOGGER.error("Request %s timed out after %ss", command, effective_timeout)
                raise

    async def _keepalive_loop(self) -> None:
        """Send periodic keepalive requests to maintain connection health."""
        try:
            while self.connected:
                await asyncio.sleep(self._keepalive_interval)

                if not self.connected:
                    break

                try:
                    _LOGGER.debug("Sending keepalive request")
                    await self.send_request(
                        "GetParamList",
                        request_timeout=10.0,  # Shorter timeout for keepalive
                        condition="OBJTYP=SYSTEM",
                        objectList=[{"objnam": "INCR", "keys": ["MODE"]}],
                    )
                except TimeoutError:
                    _LOGGER.warning("Keepalive timeout - connection may be dead")
                    await self._handle_connection_lost(ICConnectionError("Keepalive timeout"))
                    break
                except ICConnectionError as err:
                    _LOGGER.warning("Keepalive failed: %s", err)
                    await self._handle_connection_lost(err)
                    break

        except asyncio.CancelledError:
            _LOGGER.debug("Keepalive task cancelled")

    async def _handle_connection_lost(self, exc: Exception | None) -> None:
        """Handle connection loss."""
        self._connected = False

        if self._disconnect_callback:
            self._disconnect_callback(exc)

        await self.disconnect()



"""Attribute definitions for Pentair IntelliCenter.

This package contains all attribute definitions organized by type:
- constants: Type constants and attribute name constants
- body: Body of water (pool/spa) attributes
- circuit: Circuit and circuit group attributes
- equipment: Pump, heater, chemistry, and sensor attributes
- schedule: Schedule attributes
- system: System, panel, module, permit, and clock attributes
- misc: Remote, valve, external instrument, and other attributes
"""

from .body import BODY_ATTRIBUTES
from .circuit import CIRCGRP_ATTRIBUTES, CIRCUIT_ATTRIBUTES
from .constants import (
    # Attribute name constants
    ACT_ATTR,
    AVAIL_ATTR,
    BODY_ATTR,
    # Object type constants
    BODY_TYPE,
    BOOST_ATTR,
    CHEM_TYPE,
    CIRCGRP_TYPE,
    CIRCUIT_ATTR,
    CIRCUIT_TYPE,
    # Status constants
    COLOR_EFFECT_SUBTYPES,
    COMUART_ATTR,
    COOLING_ATTR,
    DLY_ATTR,
    ENABLE_ATTR,
    EXTINSTR_TYPE,
    FEATR_ATTR,
    FEATR_TYPE,
    GPM_ATTR,
    GROUP_ATTR,
    HEATER_ATTR,
    HEATER_TYPE,
    HEATING_ATTR,
    HNAME_ATTR,
    HTMODE_ATTR,
    LIGHT_EFFECTS,
    LIGHT_SUBTYPES,
    LISTORD_ATTR,
    LOTMP_ATTR,
    LSTTMP_ATTR,
    MODE_ATTR,
    MODULE_TYPE,
    NORMAL_ATTR,
    # Special values
    NULL_OBJNAM,
    OBJTYP_ATTR,
    ORPTNK_ATTR,
    ORPVAL_ATTR,
    PANEL_TYPE,
    PARENT_ATTR,
    PERMIT_ATTR,
    PERMIT_TYPE,
    PHTNK_ATTR,
    PHVAL_ATTR,
    PMPCIRC_TYPE,
    PRESS_TYPE,
    PRIM_ATTR,
    PROPNAME_ATTR,
    PUMP_STATUS_OFF,
    PUMP_STATUS_ON,
    PUMP_TYPE,
    PWR_ATTR,
    QUALTY_ATTR,
    READY_ATTR,
    REMBTN_TYPE,
    REMOTE_TYPE,
    RPM_ATTR,
    SALT_ATTR,
    SCHED_TYPE,
    SEC_ATTR,
    SELECT_ATTR,
    SENSE_TYPE,
    SHOMNU_ATTR,
    SMTSRT_ATTR,
    SNAME_ATTR,
    SOURCE_ATTR,
    SPEED_ATTR,
    STATIC_ATTR,
    STATUS_ATTR,
    STATUS_OFF,
    STATUS_ON,
    SUBTYP_ATTR,
    SUPER_ATTR,
    SYSTEM_TYPE,
    SYSTIM_TYPE,
    TEMP_ATTR,
    TIME_ATTR,
    TIMOUT_ATTR,
    UPDATE_ATTR,
    USE_ATTR,
    USER_PRIVILEGES,
    VACFLO_ATTR,
    VACTIM_ATTR,
    VALVE_TYPE,
    VER_ATTR,
    VOL_ATTR,
    # Enums
    HeaterType,
)
from .equipment import (
    CHEM_ATTRIBUTES,
    HEATER_ATTRIBUTES,
    PMPCIRC_ATTRIBUTES,
    PUMP_ATTRIBUTES,
    SENSE_ATTRIBUTES,
)
from .misc import (
    EXTINSTR_ATTRIBUTES,
    FEATR_ATTRIBUTES,
    PRESS_ATTRIBUTES,
    REMBTN_ATTRIBUTES,
    REMOTE_ATTRIBUTES,
    VALVE_ATTRIBUTES,
)
from .schedule import SCHED_ATTRIBUTES
from .system import (
    MODULE_ATTRIBUTES,
    PANEL_ATTRIBUTES,
    PERMIT_ATTRIBUTES,
    SYSTEM_ATTRIBUTES,
    SYSTIM_ATTRIBUTES,
)

# Master mapping of object types to their tracked attributes
ALL_ATTRIBUTES_BY_TYPE = {
    BODY_TYPE: BODY_ATTRIBUTES,
    CHEM_TYPE: CHEM_ATTRIBUTES,
    CIRCGRP_TYPE: CIRCGRP_ATTRIBUTES,
    CIRCUIT_TYPE: CIRCUIT_ATTRIBUTES,
    EXTINSTR_TYPE: EXTINSTR_ATTRIBUTES,
    FEATR_TYPE: FEATR_ATTRIBUTES,
    HEATER_TYPE: HEATER_ATTRIBUTES,
    MODULE_TYPE: MODULE_ATTRIBUTES,
    PANEL_TYPE: PANEL_ATTRIBUTES,
    PERMIT_TYPE: PERMIT_ATTRIBUTES,
    PMPCIRC_TYPE: PMPCIRC_ATTRIBUTES,
    PRESS_TYPE: PRESS_ATTRIBUTES,
    PUMP_TYPE: PUMP_ATTRIBUTES,
    REMBTN_TYPE: REMBTN_ATTRIBUTES,
    REMOTE_TYPE: REMOTE_ATTRIBUTES,
    SCHED_TYPE: SCHED_ATTRIBUTES,
    SENSE_TYPE: SENSE_ATTRIBUTES,
    SYSTEM_TYPE: SYSTEM_ATTRIBUTES,
    SYSTIM_TYPE: SYSTIM_ATTRIBUTES,
    VALVE_TYPE: VALVE_ATTRIBUTES,
}

__all__ = [
    # Enums
    "HeaterType",
    # Status constants
    "STATUS_ON",
    "STATUS_OFF",
    "PUMP_STATUS_ON",
    "PUMP_STATUS_OFF",
    "LIGHT_SUBTYPES",
    "COLOR_EFFECT_SUBTYPES",
    "LIGHT_EFFECTS",
    # Special values
    "NULL_OBJNAM",
    # Object type constants
    "BODY_TYPE",
    "CHEM_TYPE",
    "CIRCGRP_TYPE",
    "CIRCUIT_TYPE",
    "EXTINSTR_TYPE",
    "FEATR_TYPE",
    "HEATER_TYPE",
    "MODULE_TYPE",
    "PANEL_TYPE",
    "PERMIT_TYPE",
    "PMPCIRC_TYPE",
    "PRESS_TYPE",
    "PUMP_TYPE",
    "REMBTN_TYPE",
    "REMOTE_TYPE",
    "SCHED_TYPE",
    "SENSE_TYPE",
    "SYSTEM_TYPE",
    "SYSTIM_TYPE",
    "VALVE_TYPE",
    # Attribute name constants
    "ACT_ATTR",
    "AVAIL_ATTR",
    "BODY_ATTR",
    "BOOST_ATTR",
    "CIRCUIT_ATTR",
    "COMUART_ATTR",
    "COOLING_ATTR",
    "DLY_ATTR",
    "ENABLE_ATTR",
    "FEATR_ATTR",
    "GPM_ATTR",
    "GROUP_ATTR",
    "HEATER_ATTR",
    "HEATING_ATTR",
    "HNAME_ATTR",
    "HTMODE_ATTR",
    "LISTORD_ATTR",
    "LOTMP_ATTR",
    "LSTTMP_ATTR",
    "MODE_ATTR",
    "NORMAL_ATTR",
    "OBJTYP_ATTR",
    "ORPTNK_ATTR",
    "ORPVAL_ATTR",
    "PARENT_ATTR",
    "PERMIT_ATTR",
    "PHTNK_ATTR",
    "PHVAL_ATTR",
    "PRIM_ATTR",
    "PROPNAME_ATTR",
    "PWR_ATTR",
    "QUALTY_ATTR",
    "READY_ATTR",
    "RPM_ATTR",
    "SALT_ATTR",
    "SEC_ATTR",
    "SELECT_ATTR",
    "SHOMNU_ATTR",
    "SMTSRT_ATTR",
    "SNAME_ATTR",
    "SOURCE_ATTR",
    "SPEED_ATTR",
    "STATIC_ATTR",
    "STATUS_ATTR",
    "SUBTYP_ATTR",
    "SUPER_ATTR",
    "TEMP_ATTR",
    "TIME_ATTR",
    "TIMOUT_ATTR",
    "UPDATE_ATTR",
    "USE_ATTR",
    "USER_PRIVILEGES",
    "VACFLO_ATTR",
    "VACTIM_ATTR",
    "VER_ATTR",
    "VOL_ATTR",
    # Attribute sets
    "ALL_ATTRIBUTES_BY_TYPE",
    "BODY_ATTRIBUTES",
    "CHEM_ATTRIBUTES",
    "CIRCGRP_ATTRIBUTES",
    "CIRCUIT_ATTRIBUTES",
    "EXTINSTR_ATTRIBUTES",
    "FEATR_ATTRIBUTES",
    "HEATER_ATTRIBUTES",
    "MODULE_ATTRIBUTES",
    "PANEL_ATTRIBUTES",
    "PERMIT_ATTRIBUTES",
    "PMPCIRC_ATTRIBUTES",
    "PRESS_ATTRIBUTES",
    "PUMP_ATTRIBUTES",
    "REMBTN_ATTRIBUTES",
    "REMOTE_ATTRIBUTES",
    "SCHED_ATTRIBUTES",
    "SENSE_ATTRIBUTES",
    "SYSTEM_ATTRIBUTES",
    "SYSTIM_ATTRIBUTES",
    "VALVE_ATTRIBUTES",
]



"""Type and attribute name constants for Pentair IntelliCenter."""

from enum import IntEnum

# Special object name for null/empty references
NULL_OBJNAM = "00000"

# Status values for different object types
STATUS_ON = "ON"
STATUS_OFF = "OFF"
PUMP_STATUS_ON = "10"
PUMP_STATUS_OFF = "4"

# Light subtypes that represent illumination devices
LIGHT_SUBTYPES = frozenset(["LIGHT", "INTELLI", "GLOW", "GLOWT", "DIMMER", "MAGIC2"])

# Light subtypes that support color effects
COLOR_EFFECT_SUBTYPES = frozenset(["INTELLI", "MAGIC2", "GLOW"])

# Mapping of IntelliCenter color effect codes to human-readable names
# Used with the USE attribute on lights with COLOR_EFFECT_SUBTYPES
LIGHT_EFFECTS: dict[str, str] = {
    "PARTY": "Party Mode",
    "CARIB": "Caribbean",
    "SSET": "Sunset",
    "ROMAN": "Romance",
    "AMERCA": "American",
    "ROYAL": "Royal",
    "WHITER": "White",
    "REDR": "Red",
    "BLUER": "Blue",
    "GREENR": "Green",
    "MAGNTAR": "Magenta",
}


class HeaterType(IntEnum):
    """Heater mode types for SetHeatMode command.

    These values correspond to the MODE parameter used by IntelliCenter
    to control heater operation.
    """

    NO_CHANGE = 0
    OFF = 1
    HEATER = 2
    SOLAR_ONLY = 3
    SOLAR_PREFERRED = 4
    ULTRA_TEMP = 5
    ULTRA_TEMP_PREFERRED = 6
    HYBRID_GAS = 7
    HYBRID_ULTRA_TEMP = 8
    HYBRID_HYBRID = 9
    HYBRID_DUAL = 10
    MASTER_TEMP = 11
    MAX_E_THERM = 12
    ETI_250 = 13


# Object type constants
BODY_TYPE = "BODY"
CHEM_TYPE = "CHEM"
CIRCUIT_TYPE = "CIRCUIT"
CIRCGRP_TYPE = "CIRCGRP"
EXTINSTR_TYPE = "EXTINSTR"
FEATR_TYPE = "FEATR"
HEATER_TYPE = "HEATER"
MODULE_TYPE = "MODULE"
PANEL_TYPE = "PANEL"
PERMIT_TYPE = "PERMIT"
PMPCIRC_TYPE = "PMPCIRC"
PRESS_TYPE = "PRESS"
PUMP_TYPE = "PUMP"
REMBTN_TYPE = "REMBTN"
REMOTE_TYPE = "REMOTE"
SCHED_TYPE = "SCHED"
SENSE_TYPE = "SENSE"
SYSTEM_TYPE = "SYSTEM"
SYSTIM_TYPE = "SYSTIM"
VALVE_TYPE = "VALVE"

# Attribute name constants (alphabetically sorted)
ACT_ATTR = "ACT"
AVAIL_ATTR = "AVAIL"
BODY_ATTR = "BODY"
BOOST_ATTR = "BOOST"
CIRCUIT_ATTR = "CIRCUIT"
COMUART_ATTR = "COMUART"
COOLING_ATTR = "COOLING"
DLY_ATTR = "DLY"
ENABLE_ATTR = "ENABLE"
FEATR_ATTR = "FEATR"
GPM_ATTR = "GPM"
GROUP_ATTR = "GROUP"
HEATER_ATTR = "HEATER"
HEATING_ATTR = "HEATING"
HNAME_ATTR = "HNAME"
HTMODE_ATTR = "HTMODE"
LISTORD_ATTR = "LISTORD"
LOTMP_ATTR = "LOTMP"
LSTTMP_ATTR = "LSTTMP"
MODE_ATTR = "MODE"
NORMAL_ATTR = "NORMAL"
OBJTYP_ATTR = "OBJTYP"
ORPTNK_ATTR = "ORPTNK"
ORPVAL_ATTR = "ORPVAL"
PARENT_ATTR = "PARENT"
PERMIT_ATTR = "PERMIT"
PHTNK_ATTR = "PHTNK"
PHVAL_ATTR = "PHVAL"
PRIM_ATTR = "PRIM"
PROPNAME_ATTR = "PROPNAME"
PWR_ATTR = "PWR"
QUALTY_ATTR = "QUALTY"
READY_ATTR = "READY"
RPM_ATTR = "RPM"
SALT_ATTR = "SALT"
SEC_ATTR = "SEC"
SELECT_ATTR = "SELECT"
SHOMNU_ATTR = "SHOMNU"
SMTSRT_ATTR = "SMTSRT"
SNAME_ATTR = "SNAME"
SOURCE_ATTR = "SOURCE"
SPEED_ATTR = "SPEED"
STATIC_ATTR = "STATIC"
STATUS_ATTR = "STATUS"
SUBTYP_ATTR = "SUBTYP"
SUPER_ATTR = "SUPER"
TEMP_ATTR = "TEMP"
TIME_ATTR = "TIME"
TIMOUT_ATTR = "TIMOUT"
UPDATE_ATTR = "UPDATE"
USE_ATTR = "USE"
VACFLO_ATTR = "VACFLO"
VACTIM_ATTR = "VACTIM"
VER_ATTR = "VER"
VOL_ATTR = "VOL"

# User privileges mapping
USER_PRIVILEGES = {
    "p": "Pool Access",
    "P": "Pool temperature",
    "h": "Pool Heat Mode",
    "m": "Spa Access",
    "S": "Spa Temperature",
    "n": "Spa Heat Mode",
    "e": "Schedule Access",
    "v": "Vacation Mode",
    "f": "Features Access",
    "l": "Lights Access",
    "c": "Chemistry Access",
    "u": "Usage Access",
    "C": "System Configuration",
    "o": "Support",
    "q": "Alerts and Notifications",
    "i": "User Portal",
    "k": "Groups",
    "a": "Advanced Settings",
    "t": "Status",
    "x": "Service Mode Circuits",
    "g": "General Settings",
}



"""Equipment attribute definitions (pumps, heaters, chemistry, sensors)."""

from .constants import (
    BODY_ATTR,
    CIRCUIT_ATTR,
    COMUART_ATTR,
    DLY_ATTR,
    GPM_ATTR,
    HNAME_ATTR,
    HTMODE_ATTR,
    LISTORD_ATTR,
    MODE_ATTR,
    ORPTNK_ATTR,
    ORPVAL_ATTR,
    PARENT_ATTR,
    PHTNK_ATTR,
    PHVAL_ATTR,
    PRIM_ATTR,
    PWR_ATTR,
    QUALTY_ATTR,
    READY_ATTR,
    RPM_ATTR,
    SALT_ATTR,
    SEC_ATTR,
    SELECT_ATTR,
    SHOMNU_ATTR,
    SNAME_ATTR,
    SOURCE_ATTR,
    STATIC_ATTR,
    STATUS_ATTR,
    SUBTYP_ATTR,
    SUPER_ATTR,
    TIME_ATTR,
    TIMOUT_ATTR,
)

# Chemistry controller attributes (IntelliChlor, IntelliChem)
CHEM_ATTRIBUTES = {
    "ALK",  # (int) IntelliChem: Alkalinity setting
    BODY_ATTR,  # (objnam) BODY being managed
    "CALC",  # (int) IntelliChem: Calcium Harness setting
    "CHLOR",  # (ON/OFF) IntelliChem: ??
    COMUART_ATTR,  # (int) X25 related ?
    "CYACID",  # (int) IntelliChem: Cyanuric Acid setting
    LISTORD_ATTR,  # (int) used to order in UI
    "ORPHI",  # (ON/OFF) IntelliChem: ORP Level too high?
    "ORPLO",  # (ON/OFF) IntelliChem: ORP Level too low?
    "ORPSET",  # (int) IntelliChem ORP level setting
    ORPTNK_ATTR,  # (int) IntelliChem: ORP Tank Level
    ORPVAL_ATTR,  # (int) IntelliChem: ORP Level
    "PHHI",  # (ON/OFF) IntelliChem: Ph Level too low?
    "PHLO",  # (ON/OFF) IntelliChem: Ph Level too low?
    "PHSET",  # (float) IntelliChem Ph level setting
    PHTNK_ATTR,  # (int) IntelliChem: Ph Tank Level
    PHVAL_ATTR,  # (float) IntelliChem: Ph Level
    PRIM_ATTR,  # (int) IntelliChlor: primary body output setting in %
    QUALTY_ATTR,  # (float) IntelliChem: Water Quality (Saturation Index)
    SALT_ATTR,  # (int) Salt level
    SEC_ATTR,  # (int) IntelliChlor: secondary body output setting in %
    "SHARE",  # (objnam) ??
    "SINDEX",  # (int) ??
    SNAME_ATTR,  # friendly name
    SUBTYP_ATTR,  # 'ICHLOR' for IntelliChlor, 'ICHEM' for IntelliChem
    SUPER_ATTR,  # (ON/OFF) IntelliChlor: turn on Boost mode (aka Super Chlorinate)
    TIMOUT_ATTR,  # (int) IntelliChlor: in seconds ??
}

# Heater attributes
# Matches node-intellicenter GetHeaters attributes
HEATER_ATTRIBUTES = {
    BODY_ATTR,  # the objnam of the body the heater serves or a list (separated by a space)
    "BOOST",  # (int) Boost mode setting
    COMUART_ATTR,  # X25 related?
    "COOL",  # (ON/OFF) Cooling mode
    DLY_ATTR,  # (int) Delay setting
    "HEATING",  # (ON/OFF) Currently heating
    HNAME_ATTR,  # equals to OBJNAM
    HTMODE_ATTR,  # (int) Heat mode setting
    LISTORD_ATTR,  # (int) used to order in UI
    MODE_ATTR,  # (int) Current operating mode (see HeaterType enum)
    PARENT_ATTR,  # (objnam) parent (module) for this heater
    "PERMIT",  # (str) Permissions
    READY_ATTR,  # (ON/OFF) Ready state
    SHOMNU_ATTR,  # (str) Menu permissions
    SNAME_ATTR,  # (str) Friendly name
    "START",  # (int) Start time
    STATIC_ATTR,  # (ON/OFF) Static setting
    STATUS_ATTR,  # (ON/OFF) Only seen 'ON'
    "STOP",  # (int) Stop time
    SUBTYP_ATTR,  # type of heater 'GENERIC','SOLAR','ULTRA','HEATER'
    TIME_ATTR,  # (int) Time setting
    TIMOUT_ATTR,  # (int) Timeout setting
}

# Pump attributes
PUMP_ATTRIBUTES = {
    BODY_ATTR,  # the objnam of the body the pump serves or a list (separated by a space)
    CIRCUIT_ATTR,  # (int) ??? only seen 1
    COMUART_ATTR,  # X25 related?
    HNAME_ATTR,  # same as objnam
    GPM_ATTR,  # (int) when applicable, real time Gallon Per Minute
    LISTORD_ATTR,  # (int) used to order in UI
    "MAX",  # (int) maximum RPM
    "MAXF",  # (int) maximum GPM (if applicable, 0 otherwise)
    "MIN",  # (int) minimum RPM
    "MINF",  # (int) minimum GPM (if applicable, 0 otherwise)
    "NAME",  # seems to equal OBJNAM
    "OBJLIST",  # ([ objnam] ) a list of PMPCIRC settings
    "PRIMFLO",  # (int) Priming Speed
    "PRIMTIM",  # (int) Priming Time in minutes
    "PRIOR",  # (int) ???
    PWR_ATTR,  # (int) when applicable, real time Power usage in Watts
    RPM_ATTR,  # (int) when applicable, real time Rotation Per Minute
    "SETTMP",  # (int) Step size for RPM
    "SETTMPNC",  # (int) ???
    SNAME_ATTR,  # friendly name
    STATUS_ATTR,  # only seen 10 for on, 4 for off
    SUBTYP_ATTR,  # type of pump: 'SPEED' (variable speed), 'FLOW' (variable flow), 'VSF' (both)
    "SYSTIM",  # (int) ???
}

# Pump circuit setting attributes
PMPCIRC_ATTRIBUTES = {
    BODY_ATTR,  # not sure, I've only see '00000'
    CIRCUIT_ATTR,  # (objnam) the circuit this setting is for
    GPM_ATTR,  # (int): the flow setting for the pump if select is GPM
    LISTORD_ATTR,  # (int) used to order in UI
    PARENT_ATTR,  # (objnam) the pump the setting belongs to
    "SPEED",  # (int): the speed setting for the pump if select is RPM
    SELECT_ATTR,  # 'RPM' or 'GPM'
}

# Sensor attributes
SENSE_ATTRIBUTES = {
    "CALIB",  # (int) calibration value
    HNAME_ATTR,  # same as objnam
    LISTORD_ATTR,  # number likely used to order things in UI
    MODE_ATTR,  # I've only seen 'OFF' so far
    "NAME",  # I've only seen '00000'
    PARENT_ATTR,  # the parent's objnam
    "PROBE",  # the uncalibrated reading of the sensor
    SNAME_ATTR,  # friendly name
    SOURCE_ATTR,  # the calibrated reading of the sensor
    STATIC_ATTR,  # (ON/OFF) not sure, only seen 'ON'
    STATUS_ATTR,  # I've only seen 'OK' so far
    SUBTYP_ATTR,  # 'SOLAR','POOL' (for water), 'AIR'
}



"""mDNS discovery for Pentair IntelliCenter units.

This module provides discovery functionality to find IntelliCenter units
on the local network using mDNS (Zeroconf).

Example:
    ```python
    import asyncio
    from pyintellicenter.discovery import discover_intellicenter_units

    async def main():
        units = await discover_intellicenter_units(timeout=5.0)
        for unit in units:
            print(f"Found: {unit.name} at {unit.host}:{unit.port}")

    asyncio.run(main())
    ```
"""

from __future__ import annotations

import asyncio
import contextlib
import logging
from dataclasses import dataclass

from zeroconf import ServiceBrowser, ServiceInfo, Zeroconf
from zeroconf.asyncio import AsyncZeroconf

_LOGGER = logging.getLogger(__name__)

# IntelliCenter service type for mDNS discovery
INTELLICENTER_SERVICE_TYPE = "_http._tcp.local."
INTELLICENTER_SERVICE_NAME_PREFIX = "Pentair"

# Default discovery timeout
DEFAULT_DISCOVERY_TIMEOUT = 10.0


@dataclass(frozen=True)
class ICUnit:
    """Represents a discovered IntelliCenter unit.

    Attributes:
        name: The unit name from mDNS
        host: IP address or hostname
        port: Raw TCP port for protocol communication (typically 6681)
        ws_port: WebSocket port from mDNS (typically 6680)
        model: Model information if available
    """

    name: str
    host: str
    port: int
    ws_port: int
    model: str | None = None

    def __repr__(self) -> str:
        return f"ICUnit(name={self.name!r}, host={self.host!r}, port={self.port}, ws_port={self.ws_port})"


class ICDiscoveryListener:
    """Listener for IntelliCenter mDNS service discovery.

    Uses a queue to communicate between zeroconf's sync callbacks
    and async processing, avoiding deprecated event loop patterns.
    """

    def __init__(self, queue: asyncio.Queue[tuple[str, str, str]]) -> None:
        self._queue = queue
        self._units: dict[str, ICUnit] = {}

    @property
    def units(self) -> list[ICUnit]:
        """Return list of discovered units."""
        return list(self._units.values())

    def add_service(
        self,
        zc: Zeroconf,  # noqa: ARG002
        service_type: str,
        name: str,
    ) -> None:
        """Called when a service is discovered (sync, from zeroconf thread)."""
        # Queue for async processing - thread-safe
        try:
            self._queue.put_nowait(("add", service_type, name))
        except asyncio.QueueFull:
            _LOGGER.warning("Discovery queue full, dropping service: %s", name)

    def remove_service(
        self,
        zc: Zeroconf,  # noqa: ARG002
        service_type: str,  # noqa: ARG002
        name: str,
    ) -> None:
        """Called when a service is removed (sync, from zeroconf thread)."""
        if name in self._units:
            del self._units[name]

    def update_service(
        self,
        zc: Zeroconf,  # noqa: ARG002
        service_type: str,
        name: str,
    ) -> None:
        """Called when a service is updated (sync, from zeroconf thread)."""
        try:
            self._queue.put_nowait(("update", service_type, name))
        except asyncio.QueueFull:
            _LOGGER.warning("Discovery queue full, dropping update: %s", name)

    def add_unit(self, name: str, unit: ICUnit) -> None:
        """Add a discovered unit."""
        self._units[name] = unit


def _is_intellicenter(name: str, info: ServiceInfo) -> bool:
    """Check if the service is an IntelliCenter unit."""
    # Check name prefix
    if name.lower().startswith("pentair") or "intellicenter" in name.lower():
        return True

    # Check properties
    if info.properties:
        for key, value in info.properties.items():
            key_str = key.decode("utf-8", errors="ignore").lower()
            if value is not None:
                value_str = value.decode("utf-8", errors="ignore").lower()
                if "pentair" in value_str or "intellicenter" in value_str:
                    return True
            if "pentair" in key_str or "intellicenter" in key_str:
                return True

    return False


async def _process_discovery_queue(
    queue: asyncio.Queue[tuple[str, str, str]],
    listener: ICDiscoveryListener,
    aiozc: AsyncZeroconf,
    discovery_timeout: float,
) -> None:
    """Process discovery events from the queue."""
    end_time = asyncio.get_running_loop().time() + discovery_timeout

    while True:
        remaining = end_time - asyncio.get_running_loop().time()
        if remaining <= 0:
            break

        try:
            action, service_type, name = await asyncio.wait_for(
                queue.get(), timeout=min(remaining, 1.0)
            )
        except TimeoutError:
            continue

        if action in ("add", "update"):
            await _resolve_service(listener, aiozc, service_type, name)


async def _check_tcp_port(host: str, port: int) -> bool:
    """Check if a TCP port is connectable."""
    try:
        _, writer = await asyncio.wait_for(
            asyncio.open_connection(host, port),
            timeout=2.0,
        )
        writer.close()
        await writer.wait_closed()
        return True
    except (OSError, TimeoutError):
        return False


async def _resolve_service(
    listener: ICDiscoveryListener,
    aiozc: AsyncZeroconf,
    service_type: str,
    name: str,
) -> None:
    """Resolve service info and add to listener if it's an IntelliCenter."""
    try:
        info = await aiozc.async_get_service_info(service_type, name, timeout=3000)
        if info is None:
            return

        # Check if this is a Pentair/IntelliCenter device
        service_name = info.name or name
        if not _is_intellicenter(service_name, info):
            return

        # Extract address
        addresses = info.parsed_addresses()
        if not addresses:
            return

        host = addresses[0]
        # mDNS advertises WebSocket port (typically 6680)
        ws_port = info.port or 6680
        # Raw TCP port is WebSocket port + 1 (typically 6681)
        tcp_port = ws_port + 1

        # Verify TCP port is accessible
        if not await _check_tcp_port(host, tcp_port):
            _LOGGER.warning(
                "IntelliCenter %s found at %s but TCP port %d not accessible",
                service_name,
                host,
                tcp_port,
            )
            return

        # Extract model from properties if available
        model = None
        if info.properties:
            model_bytes = info.properties.get(b"model")
            if model_bytes is not None:
                model = model_bytes.decode("utf-8", errors="ignore")

        unit = ICUnit(
            name=service_name, host=host, port=tcp_port, ws_port=ws_port, model=model or None
        )
        listener.add_unit(name, unit)
        _LOGGER.debug(
            "Discovered IntelliCenter: %s at %s (TCP:%d, WS:%d)",
            service_name,
            host,
            tcp_port,
            ws_port,
        )

    except Exception:
        _LOGGER.exception("Error resolving service %s", name)


async def discover_intellicenter_units(
    discovery_timeout: float = DEFAULT_DISCOVERY_TIMEOUT,
) -> list[ICUnit]:
    """Discover IntelliCenter units on the local network.

    Uses mDNS (Zeroconf) to find IntelliCenter units broadcasting
    on the local network.

    Args:
        discovery_timeout: How long to wait for discovery (seconds)

    Returns:
        List of discovered ICUnit instances
    """
    # Queue for thread-safe communication between zeroconf callbacks and async code
    queue: asyncio.Queue[tuple[str, str, str]] = asyncio.Queue(maxsize=100)

    aiozc = AsyncZeroconf()
    listener = ICDiscoveryListener(queue)
    browsers: list[ServiceBrowser] = []

    try:
        # Browse for HTTP services (IntelliCenter uses HTTP on port 6681)
        browser = ServiceBrowser(
            aiozc.zeroconf,
            INTELLICENTER_SERVICE_TYPE,
            listener,  # type: ignore[arg-type]
        )
        browsers.append(browser)

        # Also try to browse for any specific IntelliCenter service types
        with contextlib.suppress(Exception):
            browser2 = ServiceBrowser(
                aiozc.zeroconf,
                "_pentair._tcp.local.",
                listener,  # type: ignore[arg-type]
            )
            browsers.append(browser2)

        # Process discovery events from the queue
        await _process_discovery_queue(queue, listener, aiozc, discovery_timeout)

        return listener.units

    finally:
        # Cancel browsers before closing zeroconf
        for browser in browsers:
            browser.cancel()
        await aiozc.async_close()


async def find_unit_by_name(
    name: str, discovery_timeout: float = DEFAULT_DISCOVERY_TIMEOUT
) -> ICUnit | None:
    """Find a specific IntelliCenter unit by name.

    Args:
        name: Name or partial name to search for (case-insensitive)
        discovery_timeout: How long to wait for discovery

    Returns:
        ICUnit if found, None otherwise
    """
    units = await discover_intellicenter_units(discovery_timeout)
    name_lower = name.lower()

    for unit in units:
        if name_lower in unit.name.lower():
            return unit

    return None


async def find_unit_by_host(
    host: str, discovery_timeout: float = DEFAULT_DISCOVERY_TIMEOUT
) -> ICUnit | None:
    """Find a specific IntelliCenter unit by IP address.

    Args:
        host: IP address to search for
        discovery_timeout: How long to wait for discovery

    Returns:
        ICUnit if found, None otherwise
    """
    units = await discover_intellicenter_units(discovery_timeout)

    for unit in units:
        if unit.host == host:
            return unit

    return None



"""Model class for storing a Pentair system.

This module provides the data model classes for representing pool equipment
and their state. It's used by the controller to maintain a synchronized
view of the IntelliCenter system.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from .attributes import (
    ALL_ATTRIBUTES_BY_TYPE,
    CIRCUIT_TYPE,
    COLOR_EFFECT_SUBTYPES,
    FEATR_ATTR,
    LIGHT_SUBTYPES,
    OBJTYP_ATTR,
    PARENT_ATTR,
    PUMP_STATUS_OFF,
    PUMP_STATUS_ON,
    PUMP_TYPE,
    SNAME_ATTR,
    STATUS_ATTR,
    STATUS_OFF,
    STATUS_ON,
    SUBTYP_ATTR,
)

if TYPE_CHECKING:
    from collections.abc import Iterator, KeysView, ValuesView

    from .types import ObjectEntry

_LOGGER = logging.getLogger(__name__)


class PoolObject:
    """Representation of an object in the Pentair system.

    A PoolObject represents a single piece of equipment (pump, light, heater, etc.)
    with its type, subtype, and current attribute values.
    """

    __slots__ = ("_objnam", "_objtype", "_subtype", "_properties")

    def __init__(self, objnam: str, params: dict[str, Any]) -> None:
        """Initialize from object name and parameters.

        Args:
            objnam: The unique object identifier (e.g., "PUMP01", "LIGHT1")
            params: Dictionary of object attributes including OBJTYP and optionally SUBTYP
        """
        self._objnam = objnam
        self._objtype: str = params.pop(OBJTYP_ATTR)
        self._subtype: str | None = params.pop(SUBTYP_ATTR, None)
        self._properties: dict[str, Any] = params

    @property
    def objnam(self) -> str:
        """Return the id of the object (OBJNAM)."""
        return self._objnam

    @property
    def sname(self) -> str | None:
        """Return the friendly name (SNAME)."""
        return self._properties.get(SNAME_ATTR)

    @property
    def objtype(self) -> str:
        """Return the object type."""
        return self._objtype

    @property
    def subtype(self) -> str | None:
        """Return the object subtype."""
        return self._subtype

    @property
    def status(self) -> str | None:
        """Return the object status."""
        return self._properties.get(STATUS_ATTR)

    @property
    def off_status(self) -> str:
        """Return the value of an OFF status."""
        return PUMP_STATUS_OFF if self._objtype == PUMP_TYPE else STATUS_OFF

    @property
    def on_status(self) -> str:
        """Return the value of an ON status."""
        return PUMP_STATUS_ON if self._objtype == PUMP_TYPE else STATUS_ON

    @property
    def is_a_light(self) -> bool:
        """Return True if the object is a light."""
        return self._objtype == CIRCUIT_TYPE and self._subtype in LIGHT_SUBTYPES

    @property
    def supports_color_effects(self) -> bool:
        """Return True if object is a light that supports color effects."""
        return self.is_a_light and self._subtype in COLOR_EFFECT_SUBTYPES

    @property
    def is_a_light_show(self) -> bool:
        """Return True if the object is a light show."""
        return self._objtype == CIRCUIT_TYPE and self._subtype == "LITSHO"

    @property
    def is_featured(self) -> bool:
        """Return True if the object is Featured."""
        return self._properties.get(FEATR_ATTR) == "ON"

    def __getitem__(self, key: str) -> Any:
        """Return the value for attribute 'key'."""
        return self._properties.get(key)

    def __str__(self) -> str:
        """Return a friendly string representation."""
        parts = [self._objnam]
        if self._subtype:
            parts.append(f"({self._objtype}/{self._subtype}):")
        else:
            parts.append(f"({self._objtype}):")

        for key in sorted(self._properties):
            value = self._properties[key]
            if isinstance(value, list):
                value = "[" + ",".join(f"{{{v}}}" for v in value) + "]"
            parts.append(f"{key}: {value}")
        return " ".join(parts)

    def __repr__(self) -> str:
        """Return a detailed string representation for debugging."""
        return (
            f"PoolObject(objnam={self._objnam!r}, objtype={self._objtype!r}, "
            f"subtype={self._subtype!r}, properties={self._properties!r})"
        )

    @property
    def attribute_keys(self) -> KeysView[str]:
        """Return a view of attribute keys for this object."""
        return self._properties.keys()

    @property
    def properties(self) -> dict[str, Any]:
        """Return the properties of the object."""
        return self._properties

    def update(self, updates: dict[str, Any]) -> dict[str, Any]:
        """Update the object from a set of key/value pairs.

        Args:
            updates: Dictionary of attribute updates to apply

        Returns:
            Dictionary of attributes that actually changed
        """
        changed: dict[str, Any] = {}

        for key, value in updates.items():
            # Check if value is unchanged using single lookup
            current = self._properties.get(key)
            if current == value:
                continue

            # Handle type/subtype updates (rare but possible)
            if key == OBJTYP_ATTR:
                self._objtype = value
            elif key == SUBTYP_ATTR:
                self._subtype = value
            else:
                self._properties[key] = value
            changed[key] = value

        return changed


class PoolModel:
    """Representation of a subset of the underlying Pentair system.

    The PoolModel maintains a collection of PoolObjects and provides
    methods for querying and updating them. It filters objects based
    on an attribute map to only track relevant equipment types.
    """

    def __init__(
        self,
        attribute_map: dict[str, set[str]] | None = None,
    ) -> None:
        """Initialize the model.

        Args:
            attribute_map: Optional mapping of object types to their tracked attributes.
                         Defaults to ALL_ATTRIBUTES_BY_TYPE.
        """
        self._objects: dict[str, PoolObject] = {}
        self._attribute_map = attribute_map if attribute_map is not None else ALL_ATTRIBUTES_BY_TYPE

    @property
    def object_values(self) -> ValuesView[PoolObject]:
        """Return a view of all objects (no copy)."""
        return self._objects.values()

    @property
    def objects(self) -> dict[str, PoolObject]:
        """Return the dictionary of objects contained in the model."""
        return self._objects

    @property
    def num_objects(self) -> int:
        """Return the number of objects contained in the model."""
        return len(self._objects)

    def __iter__(self) -> Iterator[PoolObject]:
        """Allow iteration over all values."""
        return iter(self._objects.values())

    def __getitem__(self, key: str) -> PoolObject | None:
        """Return an object based on its objnam."""
        return self._objects.get(key)

    def __repr__(self) -> str:
        """Return a detailed string representation for debugging."""
        return (
            f"PoolModel(num_objects={self.num_objects}, types={list(self._attribute_map.keys())})"
        )

    def get_by_type(self, obj_type: str, subtype: str | None = None) -> list[PoolObject]:
        """Return all objects which match the type and optional subtype.

        Args:
            obj_type: The object type to filter by (e.g., 'BODY', 'PUMP')
            subtype: Optional subtype to further filter (e.g., 'SPA', 'POOL')

        Returns:
            List of matching PoolObjects

        Examples:
            get_by_type('BODY') will return all objects of type 'BODY'
            get_by_type('BODY', 'SPA') will only return the Spa
        """
        return [
            obj
            for obj in self._objects.values()
            if obj.objtype == obj_type and (subtype is None or obj.subtype == subtype)
        ]

    def get_children(self, pool_object: PoolObject) -> list[PoolObject]:
        """Return the children of a given object.

        Args:
            pool_object: The parent object

        Returns:
            List of child PoolObjects
        """
        parent_objnam = pool_object.objnam
        return [obj for obj in self._objects.values() if obj[PARENT_ATTR] == parent_objnam]

    def add_object(self, objnam: str, params: dict[str, Any]) -> PoolObject | None:
        """Update the model with a new object.

        If the object already exists, updates its attributes instead.
        Only adds objects whose type is in the attribute map.

        Args:
            objnam: The object identifier
            params: Dictionary of object attributes

        Returns:
            The created/updated PoolObject, or None if type not in attribute map
        """
        pool_obj = self._objects.get(objnam)

        if pool_obj is None:
            pool_obj = PoolObject(objnam, params)
            if pool_obj.objtype in self._attribute_map:
                self._objects[objnam] = pool_obj
            else:
                return None
        else:
            pool_obj.update(params)
        return pool_obj

    def add_objects(self, obj_list: list[ObjectEntry]) -> None:
        """Create or update from all the objects in the list.

        Args:
            obj_list: List of objects with 'objnam' and 'params' keys
        """
        for entry in obj_list:
            self.add_object(entry["objnam"], entry["params"])

    def attributes_to_track(self) -> list[dict[str, Any]]:
        """Return all the object/attributes we want to track.

        Returns:
            List of query items with 'objnam' and 'keys' for each object
        """
        query: list[dict[str, Any]] = []
        for pool_obj in self._objects.values():
            attributes = self._attribute_map.get(pool_obj.objtype)
            if attributes is None:
                # If we don't specify a set of attributes for this object type,
                # default to all known attributes for this type
                attributes = ALL_ATTRIBUTES_BY_TYPE.get(pool_obj.objtype)
            if attributes:
                query.append({"objnam": pool_obj.objnam, "keys": list(attributes)})
        return query

    def process_updates(self, updates: list[ObjectEntry]) -> dict[str, dict[str, Any]]:
        """Update the state of the objects in the model.

        Args:
            updates: List of updates with 'objnam' and 'params' keys

        Returns:
            Dictionary mapping objnam to their changed attributes
        """
        updated: dict[str, dict[str, Any]] = {}
        for update in updates:
            objnam = update["objnam"]
            pool_obj = self._objects.get(objnam)
            if pool_obj is not None:
                changed = pool_obj.update(update["params"])
                if changed:
                    updated[objnam] = changed
        return updated



"""Schedule attribute definitions."""

from .constants import (
    ACT_ATTR,
    AVAIL_ATTR,
    CIRCUIT_ATTR,
    COOLING_ATTR,
    GROUP_ATTR,
    HEATER_ATTR,
    HNAME_ATTR,
    LISTORD_ATTR,
    LOTMP_ATTR,
    MODE_ATTR,
    SMTSRT_ATTR,
    SNAME_ATTR,
    STATIC_ATTR,
    STATUS_ATTR,
    TIME_ATTR,
    TIMOUT_ATTR,
    UPDATE_ATTR,
    VACFLO_ATTR,
    VACTIM_ATTR,
)

# Schedule attributes
# Matches node-intellicenter GetSchedule attributes
SCHED_ATTRIBUTES = {
    ACT_ATTR,  # (ON/OFF) ON if schedule is currently active
    AVAIL_ATTR,  # Availability status
    CIRCUIT_ATTR,  # (objnam) The circuit controlled by this schedule
    COOLING_ATTR,  # (ON/OFF) Cooling mode for this schedule
    "DAY",  # Days this schedule runs (e.g., 'MTWRFAU' for every day, 'AU' for weekends)
    "DNTSTP",  # (ON/OFF) Don't Stop - Set to ON to never end
    GROUP_ATTR,  # Schedule group
    HEATER_ATTR,  # Set to HEATER objnam if schedule should trigger heating
    # '00000' for off, '00001' for Don't Change
    "HITMP",  # (int) High temperature setting
    HNAME_ATTR,  # Same as objnam
    LISTORD_ATTR,  # (int) Used to order in UI
    LOTMP_ATTR,  # (int) When heater is set, this is the desired temperature
    MODE_ATTR,  # (str) Schedule mode
    "SINGLE",  # (ON/OFF) ON if the schedule should not repeat
    SMTSRT_ATTR,  # Smart start setting
    SNAME_ATTR,  # Friendly name of the schedule
    "START",  # Start time mode: 'ABSTIM' (absolute), 'SRIS' (sunrise), 'SSET' (sunset)
    STATIC_ATTR,  # (ON/OFF) Static setting
    STATUS_ATTR,  # (ON/OFF) ON if schedule is active
    "STOP",  # Stop time mode: 'ABSTIME', 'SRIS', or 'SSET'
    TIME_ATTR,  # Time the schedule starts in 'HH,MM,SS' format (24h clock)
    TIMOUT_ATTR,  # Time the schedule stops in 'HH,MM,SS' format (24h clock)
    UPDATE_ATTR,  # Last update timestamp
    VACFLO_ATTR,  # (ON/OFF) ON if schedule only applies to Vacation Mode
    VACTIM_ATTR,  # Vacation time setting
}
